@@ -2,10 +2,16 @@
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 import datetime
-from itertools import groupby
+from itertools import groupby, combinations
 from operator import itemgetter
+from collections import defaultdict
+
+try:
+    from sql import Null
+except ImportError:
+    Null = None
 from sql.aggregate import Sum
-from sql.conditionals import Coalesce
+from sql.conditionals import Coalesce, Case
 
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.wizard import Wizard, StateTransition, StateView, StateAction, \
@@ -16,12 +22,16 @@ from trytond.pyson import Eval, Bool, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 from trytond.rpc import RPC
-from trytond.tools import reduce_ids
+from trytond.tools import reduce_ids, grouped_slice
+from trytond.config import config
 
 __all__ = ['Move', 'Reconciliation', 'Line', 'OpenJournalAsk',
-    'OpenJournal', 'OpenAccount', 'ReconcileLinesWriteOff', 'ReconcileLines',
-    'UnreconcileLinesStart', 'UnreconcileLines', 'OpenReconcileLinesStart',
-    'OpenReconcileLines', 'FiscalYearLine', 'FiscalYear2',
+    'OpenJournal', 'OpenAccount',
+    'ReconcileLinesWriteOff', 'ReconcileLines',
+    'UnreconcileLines',
+    'Reconcile', 'ReconcileShow',
+    'CancelMoves',
+    'FiscalYearLine', 'FiscalYear2',
     'PrintGeneralJournalStart', 'PrintGeneralJournal', 'GeneralJournal']
 __metaclass__ = PoolMeta
 
@@ -86,6 +96,9 @@ class Move(ModelSQL, ModelView):
                     'because it\'s date is outside its period.'),
                 'draft_closed_period': ('You can not set to draft move '
                     '"%(move)s" because period "%(period)s" is closed.'),
+                'period_cancel': (
+                    'The period of move "%s" is closed.\n'
+                    'Use the current period?'),
                 })
         cls._buttons.update({
                 'post': {
@@ -156,7 +169,7 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def _get_origin(cls):
         'Return list of Model names for origin Reference'
-        return ['account.fiscalyear']
+        return ['account.fiscalyear', 'account.move']
 
     @classmethod
     def get_origin(cls):
@@ -185,7 +198,9 @@ class Move(ModelSQL, ModelView):
     def check_date(self):
         if (self.date < self.period.start_date
                 or self.date > self.period.end_date):
-            self.raise_user_error('date_outside_period', (self.rec_name,))
+            self.raise_user_error('date_outside_period', {
+                        'move': self.rec_name,
+                        })
 
     @classmethod
     def check_modify(cls, moves):
@@ -286,9 +301,7 @@ class Move(ModelSQL, ModelView):
         company = User(user).company
         amounts = {}
         move2draft_lines = {}
-        for i in range(0, len(moves), cursor.IN_MAX):
-            sub_moves = moves[i:i + cursor.IN_MAX]
-            sub_move_ids = [m.id for m in sub_moves]
+        for sub_move_ids in grouped_slice([m.id for m in moves]):
             red_sql = reduce_ids(line.move, sub_move_ids)
 
             cursor.execute(*line.select(line.move,
@@ -325,8 +338,7 @@ class Move(ModelSQL, ModelView):
                 (draft_moves, 'draft'),
                 ):
             if move_ids:
-                for i in range(0, len(move_ids), cursor.IN_MAX):
-                    sub_ids = move_ids[i:i + cursor.IN_MAX]
+                for sub_ids in grouped_slice(move_ids):
                     red_sql = reduce_ids(line.move, sub_ids)
                     # Use SQL to prevent double validate loop
                     cursor.execute(*line.update(
@@ -334,6 +346,50 @@ class Move(ModelSQL, ModelView):
                             values=[state],
                             where=red_sql))
 
+    def _cancel_default(self):
+        'Return default dictionary to cancel move'
+        pool = Pool()
+        Date = pool.get('ir.date')
+        Period = pool.get('account.period')
+
+        default = {
+            'origin': str(self),
+            }
+        if self.period.state == 'close':
+            self.raise_user_warning('%s.cancel' % self,
+                'period_cancel', self.rec_name)
+            date = Date.today()
+            period_id = Period.find(self.company.id, date=date)
+            default.update({
+                    'date': date,
+                    'period': period_id,
+                    })
+        return default
+
+    def cancel(self):
+        'Return a cancel move'
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        TaxLine = pool.get('account.tax.line')
+        default = self._cancel_default()
+        cancel_move, = self.copy([self], default=default)
+        lines = []
+        tax_lines = []
+        for line in cancel_move.lines:
+            line.debit *= -1
+            line.credit *= -1
+            lines.extend(([line], line._save_values))
+            line._values = None
+            for tax_line in line.tax_lines:
+                tax_line.amount *= -1
+                tax_lines.extend(([tax_line], tax_line._save_values))
+                tax_line._values = None
+        if lines:
+            Line.write(*lines)
+        if tax_lines:
+            TaxLine.write(*tax_lines)
+        return cancel_move
+
     @classmethod
     @ModelView.button
     def post(cls, moves):
@@ -363,11 +419,13 @@ class Move(ModelSQL, ModelView):
                     move.period.post_move_sequence_used.id)
             cls.write([move], values)
 
+            keyfunc = lambda l: (l.party, l.account)
             to_reconcile = [l for l in move.lines
                 if ((l.debit == l.credit == Decimal('0'))
                     and l.account.reconcile)]
-            if to_reconcile:
-                Line.reconcile(to_reconcile)
+            to_reconcile = sorted(to_reconcile, key=keyfunc)
+            for _, zero_lines in groupby(to_reconcile, keyfunc):
+                Line.reconcile(list(zero_lines))
 
     @classmethod
     @ModelView.button
@@ -444,7 +502,8 @@ class Reconciliation(ModelSQL, ModelView):
             debit = Decimal('0.0')
             credit = Decimal('0.0')
             account = None
-            party = None
+            if reconciliation.lines:
+                party = reconciliation.lines[0].party
             for line in reconciliation.lines:
                 if line.state != 'valid':
                     cls.raise_user_error('reconciliation_line_not_valid',
@@ -465,9 +524,7 @@ class Reconciliation(ModelSQL, ModelView):
                             'line': line.rec_name,
                             'account': line.account.rec_name,
                             })
-                if not party:
-                    party = line.party
-                elif line.party and party.id != line.party.id:
+                if line.party != party:
                     cls.raise_user_error('reconciliation_different_parties', {
                             'line': line.rec_name,
                             'party1': line.party.rec_name,
@@ -536,7 +593,14 @@ class Line(ModelSQL, ModelView):
             'required': Bool(Eval('amount_second_currency')),
             },
         depends=['amount_second_currency'])
-    party = fields.Many2One('party.party', 'Party', select=True)
+    party = fields.Many2One('party.party', 'Party', select=True,
+        states={
+            'required': Eval('party_required', False),
+            'invisible': ~Eval('party_required', False),
+            },
+        depends=['party_required'])
+    party_required= fields.Function(fields.Boolean('Party Required'),
+        'on_change_with_party_required')
     maturity_date = fields.Date('Maturity Date',
         help='This field is used for payable and receivable lines. \n'
         'You can put the limit date for the payment.')
@@ -559,7 +623,10 @@ class Line(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Line, cls).__setup__()
-        cls._check_modify_exclude = ['reconciliation']
+        cls._check_modify_exclude = {'reconciliation'}
+        cls._reconciliation_modify_disallow = {
+            'account', 'debit', 'credit', 'party',
+            }
         cls._sql_constraints += [
             ('credit_debit',
                 'CHECK(credit * debit = 0.0)',
@@ -584,6 +651,8 @@ class Line(ModelSQL, ModelView):
                 'move_inactive_account': ('You can not create a move line '
                     'with account "%s" because it is inactive.'),
                 'already_reconciled': 'Line "%s" (%d) already reconciled.',
+                'party_required': 'Party is required on line "%s"',
+                'party_set': 'Party must not be set on line "%s"',
                 })
 
     @classmethod
@@ -652,14 +721,14 @@ class Line(ModelSQL, ModelView):
         return Decimal(0)
 
     @classmethod
-    def default_get(cls, fields, with_rec_name=True, with_on_change=True):
+    def default_get(cls, fields, with_rec_name=True):
         pool = Pool()
         Move = pool.get('account.move')
         Tax = pool.get('account.tax')
         Account = pool.get('account.account')
         TaxCode = pool.get('account.tax.code')
         values = super(Line, cls).default_get(fields,
-                with_rec_name=with_rec_name, with_on_change=with_on_change)
+                with_rec_name=with_rec_name)
 
         if 'move' not in fields:
             #Not manual entry
@@ -891,6 +960,12 @@ class Line(ModelSQL, ModelView):
                     self.account.second_currency.digits
         return changes
 
+    @fields.depends('account')
+    def on_change_with_party_required(self, name=None):
+        if self.account:
+            return self.account.party_required
+        return False
+
     def _compute_tax_lines(self, journal_type):
         res = {}
         pool = Pool()
@@ -1042,17 +1117,15 @@ class Line(ModelSQL, ModelView):
         return res
 
     def get_move_field(self, name):
+        field = getattr(self.__class__, name)
         if name.startswith('move_'):
             name = name[5:]
-        if name in ('date', 'state', 'description'):
-            return getattr(self.move, name)
-        elif name == 'origin':
-            origin = getattr(self.move, name)
-            if origin:
-                return str(origin)
-            return None
-        else:
-            return getattr(self.move, name).id
+        value = getattr(self.move, name)
+        if isinstance(value, ModelSQL):
+            if field._type == 'reference':
+                return str(value)
+            return value.id
+        return value
 
     @classmethod
     def set_move_field(cls, lines, name, value):
@@ -1092,6 +1165,16 @@ class Line(ModelSQL, ModelView):
     order_origin = _order_move_field('origin')
     order_move_state = _order_move_field('state')
 
+    def get_rec_name(self, name):
+        if self.debit > self.credit:
+            return self.account.rec_name
+        else:
+            return '(%s)' % self.account.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('account.rec_name',) + tuple(clause[1:])]
+
     @classmethod
     def query_get(cls, table):
         '''
@@ -1198,6 +1281,9 @@ class Line(ModelSQL, ModelView):
         if not self.account.active:
             self.raise_user_error('move_inactive_account', (
                     self.account.rec_name,))
+        if bool(self.party) != self.account.party_required:
+            error = 'party_set' if self.party else 'party_required'
+            self.raise_user_error(error, self.rec_name)
 
     @classmethod
     def check_journal_period_modify(cls, period, journal):
@@ -1223,28 +1309,38 @@ class Line(ModelSQL, ModelView):
                         }])
 
     @classmethod
-    def check_modify(cls, lines):
+    def check_modify(cls, lines, modified_fields=None):
         '''
         Check if the lines can be modified
         '''
+        if (modified_fields is not None
+                and modified_fields <= cls._check_modify_exclude):
+            return
         journal_period_done = []
         for line in lines:
             if line.move.state == 'posted':
                 cls.raise_user_error('modify_posted_move', (
                         line.move.rec_name,))
-            if line.reconciliation:
-                cls.raise_user_error('modify_reconciled', (
-                        line.rec_name,))
             journal_period = (line.journal.id, line.period.id)
             if journal_period not in journal_period_done:
                 cls.check_journal_period_modify(line.period,
                         line.journal)
                 journal_period_done.append(journal_period)
 
+    @classmethod
+    def check_reconciliation(cls, lines, modified_fields=None):
+        if (modified_fields is not None
+                and not modified_fields & cls._reconciliation_modify_disallow):
+            return
+        for line in lines:
+            if line.reconciliation:
+                cls.raise_user_error('modify_reconciled', line.rec_name)
+
     @classmethod
     def delete(cls, lines):
         Move = Pool().get('account.move')
         cls.check_modify(lines)
+        cls.check_reconciliation(lines)
         moves = [x.move for x in lines]
         super(Line, cls).delete(lines)
         Move.validate_move(moves)
@@ -1258,8 +1354,8 @@ class Line(ModelSQL, ModelView):
         moves = []
         all_lines = []
         for lines, values in zip(actions, actions):
-            if any(k not in cls._check_modify_exclude for k in values):
-                cls.check_modify(lines)
+            cls.check_modify(lines, set(values.keys()))
+            cls.check_reconciliation(lines, set(values.keys()))
             moves.extend((x.move for x in lines))
             all_lines.extend(lines)
             args.extend((lines, values))
@@ -1380,11 +1476,14 @@ class Line(ModelSQL, ModelView):
 
         lines = lines[:]
         reconcile_account = None
+        reconcile_party = None
         amount = Decimal('0.0')
         for line in lines:
             amount += line.debit - line.credit
             if not reconcile_account:
                 reconcile_account = line.account
+            if not reconcile_party:
+                reconcile_party = line.party
         amount = reconcile_account.currency.round(amount)
         if not account and journal:
             if amount >= 0:
@@ -1403,12 +1502,18 @@ class Line(ModelSQL, ModelView):
                         'lines': [
                             ('create', [{
                                         'account': reconcile_account.id,
+                                        'party': (reconcile_party.id
+                                            if reconcile_party else None),
                                         'debit': (amount < Decimal('0.0')
                                             and - amount or Decimal('0.0')),
                                         'credit': (amount > Decimal('0.0')
                                             and amount or Decimal('0.0')),
                                         }, {
                                         'account': account.id,
+                                        'party': (reconcile_party.id
+                                            if (account.party_required and
+                                                reconcile_party)
+                                            else None),
                                         'debit': (amount > Decimal('0.0')
                                             and amount or Decimal('0.0')),
                                         'credit': (amount < Decimal('0.0')
@@ -1619,11 +1724,6 @@ class ReconcileLines(Wizard):
         return 'end'
 
 
-class UnreconcileLinesStart(ModelView):
-    'Unreconcile Lines'
-    __name__ = 'account.move.unreconcile_lines.start'
-
-
 class UnreconcileLines(Wizard):
     'Unreconcile Lines'
     __name__ = 'account.move.unreconcile_lines'
@@ -1642,37 +1742,219 @@ class UnreconcileLines(Wizard):
         return 'end'
 
 
-class OpenReconcileLinesStart(ModelView):
-    'Open Reconcile Lines'
-    __name__ = 'account.move.open_reconcile_lines.start'
-    account = fields.Many2One('account.account', 'Account', required=True,
-            domain=[('kind', '!=', 'view'), ('reconcile', '=', True)])
-
-
-class OpenReconcileLines(Wizard):
-    'Open Reconcile Lines'
-    __name__ = 'account.move.open_reconcile_lines'
-    start = StateView('account.move.open_reconcile_lines.start',
-        'account.open_reconcile_lines_start_view_form', [
+class Reconcile(Wizard):
+    'Reconcile'
+    __name__ = 'account.reconcile'
+    start_state = 'next_'
+    next_ = StateTransition()
+    show = StateView('account.reconcile.show',
+        'account.reconcile_show_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Open', 'open_', 'tryton-ok', default=True),
+            Button('Skip', 'next_', 'tryton-go-next'),
+            Button('Reconcile', 'reconcile', 'tryton-ok', default=True),
             ])
-    open_ = StateAction('account.act_move_line_form')
+    reconcile = StateTransition()
 
-    def do_open_(self, action):
-        encoder = PYSONEncoder()
-        action['pyson_domain'] = encoder.encode([
-            ('account', '=', self.start.account.id),
+    def get_accounts(self):
+        'Return a list of account id to reconcile'
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        line = Line.__table__()
+        Account = pool.get('account.account')
+        account = Account.__table__()
+        cursor = Transaction().cursor
+
+        balance = line.debit - line.credit
+        cursor.execute(*line.join(account,
+                condition=line.account == account.id).select(
+                account.id,
+                where=(line.reconciliation == Null) & account.reconcile,
+                group_by=account.id,
+                having=(
+                    Sum(Case((balance > 0, 1), else_=0)) > 0)
+                & (Sum(Case((balance < 0, 1), else_=0)) > 0)
+                ))
+        return [a for a, in cursor.fetchall()]
+
+    def get_parties(self, account):
+        'Return a list party to reconcile for the account'
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        line = Line.__table__()
+        cursor = Transaction().cursor
+
+        balance = line.debit - line.credit
+        cursor.execute(*line.select(line.party,
+                where=(line.reconciliation == Null)
+                & (line.account == account.id),
+                group_by=line.party,
+                having=(
+                    Sum(Case((balance > 0, 1), else_=0)) > 0)
+                & (Sum(Case((balance < 0, 1), else_=0)) > 0)
+                ))
+        return [p for p, in cursor.fetchall()]
+
+    def transition_next_(self):
+
+        def next_account():
+            accounts = list(self.show.accounts)
+            if not accounts:
+                return
+            account = accounts.pop()
+            self.show.account = account
+            self.show.parties = self.get_parties(account)
+            self.show.accounts = accounts
+            return account
+
+        def next_party():
+            parties = list(self.show.parties)
+            if not parties:
+                return
+            party = parties.pop()
+            self.show.party = party
+            self.show.parties = parties
+            return party
+
+        if getattr(self.show, 'accounts', None) is None:
+            self.show.accounts = self.get_accounts()
+            if not next_account():
+                return 'end'
+        if getattr(self.show, 'parties', None) is None:
+            self.show.parties = self.get_parties(self.show.account)
+
+        while not next_party():
+            if not next_account():
+                return 'end'
+        return 'show'
+
+    def default_show(self, fields):
+        pool = Pool()
+        Date = pool.get('ir.date')
+
+        defaults = {}
+        defaults['accounts'] = [a.id for a in self.show.accounts]
+        defaults['account'] = self.show.account.id
+        defaults['parties'] = [p.id for p in self.show.parties]
+        defaults['party'] = self.show.party.id if self.show.party else None
+        defaults['currency_digits'] = self.show.account.company.currency.digits
+        defaults['lines'] = self._default_lines()
+        defaults['write_off'] = Decimal(0)
+        defaults['date'] = Date.today()
+        return defaults
+
+    def _all_lines(self):
+        'Return all lines to reconcile for the current state'
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        return Line.search([
+                ('account', '=', self.show.account.id),
+                ('party', '=',
+                    self.show.party.id if self.show.party else None),
+                ('reconciliation', '=', None),
+                ])
+
+    def _default_lines(self):
+        'Return the larger list of lines which can be reconciled'
+        currency = self.show.account.company.currency
+        chunk = config.getint('account', 'reconciliation_chunk', 10)
+        # Combination is exponential so it must be limited to small number
+        default = []
+        for lines in grouped_slice(self._all_lines(), chunk):
+            lines = list(lines)
+            best = None
+            for n in xrange(len(lines), 1, -1):
+                for comb_lines in combinations(lines, n):
+                    amount = sum((l.debit - l.credit) for l in comb_lines)
+                    if currency.is_zero(amount):
+                        best = [l.id for l in comb_lines]
+                        break
+                if best:
+                    break
+            if best:
+                default.extend(best)
+        return default
+
+    def transition_reconcile(self):
+        pool = Pool()
+        Line = pool.get('account.move.line')
+
+        if self.show.lines:
+            Line.reconcile(self.show.lines,
+                journal=self.show.journal,
+                date=self.show.date,
+                description=self.show.description)
+        return 'next_'
+
+
+class ReconcileShow(ModelView):
+    'Reconcile'
+    __name__ = 'account.reconcile.show'
+    accounts = fields.One2Many('account.account', None, 'Account',
+        readonly=True)
+    account = fields.Many2One('account.account', 'Account', readonly=True)
+    parties = fields.One2Many('party.party', None, 'Parties', readonly=True)
+    party = fields.Many2One('party.party', 'Party', readonly=True)
+    lines = fields.Many2Many('account.move.line', None, None, 'Lines',
+        domain=[
+            ('account', '=', Eval('account')),
+            ('party', '=', Eval('party')),
             ('reconciliation', '=', None),
+            ],
+        depends=['account', 'party'])
+
+    _write_off_states = {
+        'required': Bool(Eval('write_off', 0)),
+        'invisible': ~Eval('write_off', 0),
+        }
+    _write_off_depends = ['write_off']
+
+    write_off = fields.Function(fields.Numeric('Write-Off',
+            digits=(16, Eval('currency_digits', 2)),
+            states=_write_off_states,
+            depends=_write_off_depends + ['currency_digits']),
+        'on_change_with_write_off')
+    currency_digits = fields.Function(fields.Integer('Currency Digits'),
+        'on_change_with_currency_digits')
+    journal = fields.Many2One('account.journal', 'Journal',
+        states=_write_off_states, depends=_write_off_depends,
+        domain=[
+            ('type', '=', 'write-off'),
             ])
-        if self.start.account.kind == 'receivable':
-            order = [('credit', 'DESC'), ('id', 'DESC')]
-        elif self.start.account.kind == 'payable':
-            order = [('debit', 'DESC'), ('id', 'DESC')]
-        else:
-            order = None
-        action['pyson_order'] = encoder.encode(order)
-        return action, {}
+    date = fields.Date('Date',
+        states=_write_off_states, depends=_write_off_depends)
+    description = fields.Char('Description',
+        states=_write_off_states, depends=_write_off_depends)
+
+    @fields.depends('lines')
+    def on_change_with_write_off(self, name=None):
+        return sum((l.debit - l.credit) for l in self.lines)
+
+    @fields.depends('account')
+    def on_change_with_currency_digits(self, name=None):
+        return self.account.company.currency.digits
+
+
+class CancelMoves(Wizard):
+    'Cancel Moves'
+    __name__ = 'account.move.cancel'
+    start_state = 'cancel'
+    cancel = StateTransition()
+
+    def transition_cancel(self):
+        pool = Pool()
+        Move = pool.get('account.move')
+        Line = pool.get('account.move.line')
+
+        moves = Move.browse(Transaction().context['active_ids'])
+        for move in moves:
+            cancel_move = move.cancel()
+            to_reconcile = defaultdict(list)
+            for line in move.lines + cancel_move.lines:
+                if line.account.reconcile:
+                    to_reconcile[line.account].append(line)
+            for lines in to_reconcile.itervalues():
+                Line.reconcile(lines)
+        return 'end'
 
 
 class FiscalYearLine(ModelSQL):
